#ifndef BIT_BUFFER_HH
#define BIT_BUFFER_HH

#include "bit/base.hh"

namespace bit {

  /** Buffer storing and retrieving variable-width values (0-32 bits)
   * in arbitrary bit positions.  The size of the buffer must be less
   * than 2^64 bits.
   */
  class Buffer {

  public:
    /** The data vector used for internal buffer representation. */
    typedef std::vector<unsigned char> DataVector;

  public:

    /** Default constructor. 
     * \param size = initial size of the buffer in bytes
     */
    Buffer(u64 size = 0)
    {
      m_data.resize(size);
    }

    /** Swap the contents of two buffers efficiently.  Invalidates
     * pointers returned earlier by the data() method. 
     */
    void swap(Buffer &buffer)
    {
      m_data.swap(buffer.m_data);
    }
    

    /** Return the size of the buffer in bytes.  
     */
    u64 size() const
    {
      return m_data.size();
    }
    
    /** Resize the buffer.  Growing the buffer does not invalidate
     * buffer contents.  If the buffer is grown, the new values are
     * initialized to zero. 
     * \param size = the new size of the buffer in bytes
     */
    void resize(u64 size)
    {
      m_data.resize(size);
    }

    /** Read value from the buffer.
     * \param byte_pos = byte position in the buffer for the read
     * \param bit_offset = bit offset for the read (0-31 bits)
     * \param bits = number of bits to read (0-32 bits)
     * \return the value read from the buffer
     * \throw bit::invalid_argument on of the bit arguments are invalid
     * \throw bit::out_of_range accessing outside the buffer
     */
    u32 read_pos(u64 byte_pos, unsigned int bit_offset, 
                 unsigned int bits) const
    {
      if (bit_offset >= max_bits_per_value)
        throw bit::invalid_argument(
          "bit::Buffer::read_pos(): invalid bit_offset");
      if (bits > max_bits_per_value)
        throw bit::invalid_argument(
          "bit::Buffer::read_pos(): tried to read too many bits");
      if (bits == 0)
        return 0;

      // Ensure that read does not exceed the buffer
      //
      u64 last_byte_required = byte_pos + (bit_offset + bits - 1) / 8;
      if (last_byte_required >= m_data.size())
        throw bit::out_of_range("bit::Buffer::read_pos(): out_of_range");

      u64 value = 0;
      int num_bytes = (bits + bit_offset + 7) / 8;
      for (int i = 0; i < num_bytes; i++)
	value += (u64)m_data[byte_pos + i] << shift[i];
      value >>= bit_offset;
      value &= one_masks[bits];
      return (u32)value;
    }

    /** Read element at a given fixed-width element
     * \param elem = the index of the element to write
     * \param bits_per_elem = the number of bits per element (0-32 bits)
     * \throw bit::invalid_argument the bit argument is invalid
     * \throw bit::out_of_range accessing outside the buffer
     */
    u32 read_elem(u64 elem, unsigned int bits_per_elem) const
    {
      u64 bit_pos = elem * bits_per_elem;
      return read_pos(bit_pos / 8, bit_pos % 8, bits_per_elem);
    }


    /** Write value in the buffer.
     * \param byte_pos = byte position in the buffer for the write
     * \param bit_offset = bit offset for the write (0-31 bits)
     * \param value = value to write (0-32 bits)
     * \param bits = number of bits to write (0-32 bits)
     * \throw bit::invalid_argument on of the bit arguments are invalid
     * \throw bit::out_of_range accessing outside the buffer
     */
    void write_pos(u64 byte_pos, unsigned int bit_offset, u32 value, 
                   unsigned int bits)
    {
      if (bit_offset >= max_bits_per_value)
        throw bit::invalid_argument(
          "bit::Buffer::write_pos(): invalid bit_offset");
      if (bits > max_bits_per_value)
        throw bit::invalid_argument(
          "bit::Buffer::write_pos(): tried to write too many bits");
      if (bit::highest_bit(value) > bits)
        throw bit::invalid_argument(
          "bit::Buffer::write_pos(): value greater than given bits");
      if (bits == 0)
        return;

      // Ensure that write does not exceed the buffer
      //
      u64 last_byte_required = byte_pos + (bit_offset + bits - 1) / 8;
      if (last_byte_required >= m_data.size())
        throw bit::out_of_range("bit::Buffer::write_pos(): out_of_range");

      u64 zero_mask = one_masks[bits];
      zero_mask <<= bit_offset;
      zero_mask = ~zero_mask;
      u64 value_mask = value;
      value_mask <<= bit_offset;

      int num_bytes = (bits + bit_offset + 7) / 8;
      for (int i = 0; i < num_bytes; i++) {
	m_data[byte_pos + i] &= (unsigned char)(zero_mask & 0xff);
	m_data[byte_pos + i] |= (unsigned char)(value_mask & 0xff);
	zero_mask >>= 8;
	value_mask >>= 8;
      }
    }

    /** Write value in the buffer in the place of given fixed-width element
     * \param elem = the index of the element to write
     * \param bits_per_elem = the number of bits per element (0-32 bits)
     * \param value = the value to write
     * \throw bit::invalid_argument the bit argument is invalid
     * \throw bit::out_of_range accessing outside the buffer
     */
    void write_elem(u64 elem, unsigned int bits_per_elem, u32 value)
    {
      u64 bit_pos = elem * bits_per_elem;
      write_pos(bit_pos / 8, bit_pos % 8, value, bits_per_elem);
    }

    /** Const access the internal data buffer.  It is safe to write
     * and read the first size() bytes of the pointer.  Note, that
     * resize() and swap may invalidate the pointer. */
    const unsigned char *data() const
    {
      return &m_data[0];
    }

    /** Access the internal data buffer.  It is safe to write
     * and read the first size() bytes of the pointer.  Note, that
     * resize() and swap may invalidate the pointer. */
    unsigned char *data() 
    {
      return &m_data[0];
    }

  private:

    /** Buffer containing the actual bit data. */
    DataVector m_data;

  };


};

#endif /* BIT_BUFFER_HH */
