#ifndef BIT_COMPRESSEDARRAY_HH
#define BIT_COMPRESSEDARRAY_HH

#include "bit/algo.hh"
#include "bit/Array.hh"
#include "str/str.hh"

namespace bit {

  /** Compressed array of sorted values.  After the array is
   * compressed, all write operations throw bit::invalid_call.
   * \warning Compressed arrays can store only up to (2^32 - 1)
   * elements because indices are also stored in arrays, and array
   * values can not be wider than 32 bits.
   *
   * Based on the following scientific publication: B.Raj and
   * E.W.D.Whittaker.  Lossless compression of language model
   * structure and word identifiers.  Proceedings of ICASSP 2003.
   * pp. 388-391.
   */
  class CompressedArray {
  public:
  
    /** Constant cost for storing shift value and pointer to inverse
     * array to avoid compressing with small gain.  This is computed
     * as 1 byte for shift and 4 bytes for the pointer regardless of
     * the machine architecture. */ 
    static const u64 constant_size_cost = 5;

    CompressedArray()
      : m_inv_array(NULL), m_shift(0)
    {
    }

    CompressedArray(const Array &array)
      : m_array(array), m_inv_array(NULL), m_shift(0)
    {
    }

    CompressedArray(const CompressedArray &array)
      : m_array(array.m_array), m_inv_array(array.m_inv_array), 
        m_shift(array.m_shift)
    {
      if (m_inv_array != NULL)
        m_inv_array = new CompressedArray(*m_inv_array);
    }

    /** Create an array with initial size.  The initial elements are
     * guaranteed to be zero.
     *
     * \param bits_per_elem = bits_per_element (0-32 bits)
     * \param num_elems = initial number of elements in the array
     * \throw bit::invalid_argument one of the bit arguments is invalid
     */
    CompressedArray(u64 num_elems, unsigned int bits_per_elem)
      : m_array(num_elems, bits_per_elem), m_inv_array(NULL), m_shift(0)
    {
    }

    ~CompressedArray()
    {
      delete m_inv_array;
    }

    CompressedArray &operator=(const CompressedArray &array)
    {
      if (&array == this)
        return *this;
      m_array = array.m_array;

      if (m_inv_array != NULL) {
        assert(m_inv_array != array.m_inv_array);
        delete m_inv_array;
      }

      m_inv_array = array.m_inv_array;
      if (m_inv_array != NULL)
        m_inv_array = new CompressedArray(*m_inv_array);
      m_shift = array.m_shift;

      return *this;
    }

    /** Return true if compressed. */
    bool is_compressed() const {
      return m_inv_array != NULL;
    }

    /** Return the number of elements stored in the compressed array. */
    u64 num_elems() const { 
      return m_array.num_elems();
    }

    /** Return the number bits used in the compression. 
     * \throw bit::invalid_call if uncompressed
     */
    unsigned int shift() const {
      if (!m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::shift() not compressed");
      return m_shift;
    }

    /** Return the number of bytes required to store the possibly
     * recursively compressed array.
     */
    u64 compressed_size() const {
      u64 size = m_array.data_len();
      if (m_inv_array)
        size += m_inv_array->compressed_size() + constant_size_cost;
      return size;
    }

    /** Change the number of elements in the array.  The values of the
     * possible new elements are guaranteed to be zero ONLY if the
     * array has never been resized smaller.  
     * \param num_elems = the new number of elements
     * \throw bit::invalid_call if called after compression
     */
    void resize(u64 num_elems)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::resize() called after compression");
      m_array.resize(num_elems);
    }
     
    /** Change the width of the elements to given width.  All elements
     * must fit in the given width or bit::invalid_argument will be thrown and
     * the array is guaranteed to be left in the original state.
     * \param bits_per_elem = the number of bits to use per element (1-32 bits)
     * \throw bit::invalid_argument invalid bits_per_elem
     * \throw bit::invalid_call if called after compression
     */
    void set_width(int bits_per_elem)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::set() called after compression");
      m_array.set_width(bits_per_elem);
    }    

    /** Set the value of an element.  
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     * \throw bit::invalid_argument value wider than the bit-width of
     * the array or trying to write to zero-width array
     * \throw bit::out_of_range accessing outside the array
     * \throw bit::invalid_call if called after compression
     */
    void set(u64 elem, u32 value)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::set() called after compression");
      m_array.set(elem, value);
    }

    /** Set the value of an element growing the buffer if necessary.
     * If the current capacity is not enough, the capacity is doubled
     * (set to one from zero), and if that is not enough, the capacity
     * is set to (elem+1).
     *
     * \warning Reserving capacity and resizing the array happens
     * before the actual value is written to the buffer.  Thus, the
     * size and capacity of the array may change even if the write
     * throws bit::invalid_argument for too wide write.
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     * \throw bit::invalid_argument value wider than the bit-width of the array
     * \throw bit::invalid_call if called after compression
     */
    void set_grow(u64 elem, u32 value)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::set_grow() called after compression");
      m_array.set_grow(elem, value);
    }

    /** Set the value of an element growing and widening the buffer if
     * necessary.  See set_grow() for info how capacity is handled and
     * warnings about exceptions.  
     *
     * \param elem = the index of the element to set
     * \param value = the value to set
     * \throw bit::invalid_call if called after compression
     */
    void set_grow_widen(u64 elem, u32 value)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::set_grow_widen() called after compression");
      m_array.set_grow_widen(elem, value);
    }

    /** Return value of an element. 
     * \param elem = the index of the element
     * \return the value
     * \throw bit::out_of_range accessing outside the array
     */
    u32 get(u64 elem) const
    {
      if (elem >= m_array.num_elems())
        throw bit::out_of_range("bit::Array::get(): out of range");
      if (!m_inv_array)
        return m_array.get(elem);

//      u64 inv = bit::last_leq(*m_inv_array, elem, 0, m_inv_array->num_elems());
      u64 inv = m_inv_array->last_leq(elem);
      assert(inv < max_u64);
      return inv << m_shift | m_array.get(elem);
    }

    /** Compress the array.  Note that if \c shift equals the original
     * array width, the array is left in its original state.
     *
     * \param shift = the number of bits to shift the array in compression
     * \throw bit::invalid_call if compressed already
     * \throw bit::invalid_argument if shift is equal or larger than 
     * bits_per_elem
     * \throw bit::out_of_range if trying to compress an array with
     * more than (2^32 - 1) elements.
     */
    void compress(unsigned int shift)
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::compress() already compressed");

      if (shift >= m_array.bits_per_elem())
        throw bit::invalid_argument(
          "bit::CompressedArray::compress() shift too large");

      if (m_array.num_elems() > 0xffffffff)
        throw bit::out_of_range(
          "bit::CompressedArray::compress() too many elements");

      if (shift == m_array.bits_per_elem())
        return;
      if (m_array.num_elems() == 0)
        return;

//      m_inv_array = new CompressedArray(inverse(shift_array(m_array, shift)));

      // Create the inverse array
      //
      {
        Array new_array(m_array);
        u32 prev = new_array.get(0);
        for (u64 i = 0; i < new_array.num_elems(); i++) {
          u32 value = new_array.get(i);
          if (value < prev)
            throw bit::invalid_argument(
              "bit::CompressedArray::compress() array not sorted");
          prev = value;
          if (shift > 0)
            new_array.set(i, value >> shift);
        }
        m_inv_array = new CompressedArray(inverse(new_array));
      }

      m_array = mask(m_array, shift);
      m_shift = shift;
    }

    /** Compress the array once with the optimal compression width. 
     * \throw bit::invalid_call if compressed already
     * \throw bit::out_of_range if trying to compress an array with
     * more than (2^32 - 1) elements.
     */
    void optimal_compress()
    {
      if (m_inv_array)
        throw bit::invalid_call(
          "bit::CompressedArray::compress() already compressed");

      unsigned int best_shift = (unsigned int)-1;
      u64 best_size = compressed_size();
      for (unsigned int shift = 0; shift < m_array.bits_per_elem(); shift++) {
        CompressedArray array(*this);
        array.compress(shift);
        if (array.compressed_size() < best_size) {
          best_shift = shift;
          best_size = array.compressed_size();
        }
      }

      if (best_shift == (unsigned int)-1)
        return;

      if (best_shift < m_array.bits_per_elem())
        compress(best_shift);
    }

    /** Compress the array and the resulting arrays recursively with
     * the optimal compression width. 
     * \throw bit::invalid_call if compressed already
     * \throw bit::out_of_range if trying to compress an array with
     * more than (2^32 - 1) elements.
     */
    void recursive_optimal_compress()
    {
      optimal_compress();
      if (is_compressed())
        m_inv_array->recursive_optimal_compress();
    }

    /** Uncompress the array.  It is safe to call this for
     * uncompressed array. */
    void uncompress()
    {
      if (!m_inv_array)
        return;

      Array new_array;
      for (u64 i = 0; i < m_array.num_elems(); i++)
        new_array.set_grow_widen(i, get(i));
      delete m_inv_array;
      m_inv_array = NULL;
      m_array = new_array;
      m_shift = 0;
    }

    /** Find the largest index that has a equal or smaller value than
     * specified.  This function is a specialized version of the
     * general bit::last_leq() as described in Raj and Whittaker
     * (2003).
     * 
     * \param value = the candidate value
     * \return the largest index containing equal or smaller value or
     * max_u64 if no such index in array
     */
    u64 last_leq(u32 value)
    {
      if (!m_inv_array)
        return bit::last_leq(m_array, value);

      u32 s = value >> m_shift;

      if (s >= m_inv_array->num_elems()) {
        if (m_array.num_elems() == 0)
          return max_u64;
        return m_array.num_elems() - 1;
      }

      u64 first = m_inv_array->get(s);
      u64 limit = m_array.num_elems();
      if (s + 1 < m_inv_array->num_elems()) 
        limit = m_inv_array->get(s + 1);
      assert(first <= limit);
      u64 last_leq = bit::last_leq(m_array, value & one_masks[m_shift],
                                   first, limit);
      if (last_leq == max_u64) {
        if (first == 0)
          return max_u64;
        return first - 1;
      }
  
      return last_leq;
    }

    template <class A>
    static A inverse(const A &array)
    {
      A inv;
      for (int j = 0;; j++) {
        u64 bound = bit::lower_bound(array, j, 0, array.num_elems());
        if (bound == array.num_elems())
          break;
        inv.set_grow_widen(inv.num_elems(), bound);
      }
      return inv;
    }

    template <class A>
    static A shift_array(const A &array, int k)
    {
      A ret(array);
      if (k > 0) {
        for (u64 i = 0; i < array.num_elems(); i++)
          ret.set(i, ret.get(i) >> k);
      }
      ret.set_width(highest_bit(max(ret)));
      return ret;
    }

    template <class A>
    static A mask(const A &array, unsigned int k)
    {
      assert(k <= 32);
      int mask = one_masks[k];
      A ret(array);
      for (u64 i = 0; i < array.num_elems(); i++)
        ret.set(i, ret.get(i) & mask);
      ret.set_width(highest_bit(max(ret)));
      return ret;
    }

    /** Return a debug string displaying the contents of the
     * compressed array. 
     * \param indent = the number of white space indentation
     */
    std::string debug_str(int indent = 0) const
    {
      u64 i = 0;
      bool a_end = false;
      bool i_end = false;
      std::string ret;
      ret.append(indent, ' ');
      ret.append(str::fmt(256, "m_shift = %d\n", m_shift));
      ret.append(indent, ' ');
      ret.append("i\tarray\tinv\n");
      while (1) {
        
        if (i >= m_array.num_elems())
          a_end = true;
        if (!m_inv_array || i >= m_inv_array->num_elems())
          i_end = true;
        
        if (a_end && i_end)
          break;

        ret.append(indent, ' ');
        ret.append(str::fmt(64, "%d:\t", i));
        if (!a_end)
          ret.append(str::fmt(64, "%d", m_array.get(i)));
        ret.append("\t");
        if (!i_end)
          ret.append(str::fmt(64, "%d", m_inv_array->get(i)));
        ret.append("\n");

        i++;
      }
      ret.append(indent, ' ');
      ret.append(str::fmt(256, "compressed size: %d bytes\n", 
                          compressed_size()));
      
      if (m_inv_array)
        ret.append(m_inv_array->debug_str(indent + 2));

      return ret;
    }

    /** Write the array in file. 
     * \param file = file stream to write to
     * \throw bit::io_error if write fails
     */
    void write(FILE *file) const
    {
      fputs("CARRAY1:", file);
      try {
        m_array.write(file);
      }
      catch (bit::io_error &e) {
        throw bit::io_error(
          std::string("bit::CompressedArray::write() write failed") +
          e.what());
      }
      if (m_inv_array) {
        fputs("+", file);
        fprintf(file, "%d:", m_shift);
        m_inv_array->write(file);
      }
      else {
        fputs("-", file);
      }
    }

    /** Read the array from file.  \bug The array can be left in
     * corrupted state if read fails with exception.
     *
     * \param file = file stream to read from
     * \throw bit::io_error if read fails
     */
    void read(FILE *file)
    {
      int version;
      int ret = fscanf(file, "CARRAY%d:", &version);
      if (ret != 1 || version != 1)
        throw bit::io_error(
          "bit::CompressedArray::read() error while reading header");
      try {
        m_array.read(file);
      }
      catch (bit::io_error &e) {
        throw bit::io_error(
          std::string("bit::CompressedArray::read() read failed: ") 
          + e.what());
      }
      int c = fgetc(file);
      if (c == '-') {
        delete m_inv_array;
        m_inv_array = NULL;
        m_shift = 0;
      }
      else if (c == '+') {
        delete m_inv_array;
        m_inv_array = new CompressedArray();
        ret = fscanf(file, "%d:", &m_shift);
        if (ret != 1 || m_shift >= max_bits_per_value)
          throw bit::io_error(
            "bit::CompressedArray::read() file format error: invalid shift");
        m_inv_array->read(file);
      }
      else
        throw bit::io_error("bit::CompressedArray::read() file format error");
    }    

  private:

    /** The masked part of the compressed array. */
    Array m_array;
    
    /** The inverse part of the compressed array.  If NULL, then the
     * array is uncompressed, \ref m_array contains the array, and
     * \ref m_shift is not defined. */
    CompressedArray *m_inv_array;

    /** The amount of shift used in array compression. */
    unsigned int m_shift;

  };

};

#endif /* BIT_COMPRESSEDARRAY_HH */
