#ifndef BIT_TRIE_HH
#define BIT_TRIE_HH

#include "bit/algo.hh"
#include "bit/CompressedArray.hh"

namespace bit {

  /** A trie structure for storing a set of integer vectors compactly.
   * Each vector is stored in a tree as a path from the root node so
   * that common prefixes are shared.  In order to avoid storing all
   * children pointers in a node, the following structure is used.
   *
   * - Each level of the tree is stored in a separate matrix.
   *
   * - The node row in a matrix contains two elements: symbol and one
   * row index to the next level matrix.  The row index specifies the
   * first row after the children of this node.  The row containing the
   * first child (on the next level) can be read from the previous row
   * (on the current level).
   *
   * - All children of a node are stored subsequently on the next level,
   * and the symbols are sorted to allow fast access.
   *
   * - Vector (x y z) can exist only if (x y) exists.
   *
   * Miscellaneous notes:
   *
   * - Nodes can be inserted at any level as long as nodes are always
   * inserted at the end of the vector on each level.
   *
   * - If (H w) is the last string, then on the parent level, for each
   * I:I>H, the child_limit value of (I) is zero
   *
   * - Otherwise, the child_limit is the index of the last child plus one
   */
  template <class Array>
  class Trie {

  public:

    /** A class for traversing in a trie. */
    class Iterator {
    public:

      /** \name Creating iterators */
      //@{

      /** Construct an un-attached iterator. */
      Iterator()
        : m_trie(NULL)
      {
      }

      /** Construct an iterator attached to an trie.  The iterator
       * will be located at the root of the trie.  
       * \param trie = trie to which the iterator is attached
       */
      Iterator(const Trie<Array> &trie)
        : m_trie(&trie)
      {
      }

      /** Construct a prefix iterator containing only a few first
       * symbols. 
       * \param it = iterator to copy
       * \param length = the number of symbols copied
       */
      Iterator(const Iterator &it, unsigned int length)
        : m_trie(it.m_trie)
      {
        if (length > it.length())
          length = it.length();
        m_symbol_stack.resize(length);
        m_child_limit_stack.resize(length);
        for (size_t i = 0; i < length; i++) {
          m_symbol_stack[i] = it.m_symbol_stack[i];
          m_child_limit_stack[i] = it.m_child_limit_stack[i];
        }
      }
      
      //@}



      /** \name Iterator information */
      //@{

      /** The length of the iterator inside the trie.  The length is
       * zero, when the iterator is positioned at the abstract root
       * node.  
       */
      unsigned int length() const
      {
        return m_symbol_stack.size();
      }

      /** The iterator position at the symbol vector.
       * \param level = the level from which the position is fetched
       * \throw bit::out_of_range if \c level is greater than length() - 1
       */
      u32 symbol_index(unsigned int level) const
      {
        if (level >= length())
          throw bit::out_of_range(
            "bit::Trie::Iterator::symbol_index() level out of range");
        assert(!m_symbol_stack.empty());
        return m_symbol_stack[level];
      }

      /** The iterator position at the symbol vector on highest level.
       * \throw bit::out_of_range if the iterator is positioned at the
       * virtual root node
       */
      u32 symbol_index() const
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::Iterator::symbol_index() called at root node");
        return symbol_index(length() - 1);
      }

      /** The iterator position at the child_limit vector.
       * \param level = the level from which the position is fetched
       * \return max_u32 if leaf-node in separated array
       * \throw bit::out_of_range if \c level is greater than length() - 1
       */
      u32 child_limit_index(unsigned int level) const
      {
        if (level >= length())
          throw bit::out_of_range(
            "bit::Trie::Iterator::child_limit_index() level out of range");
        assert(!m_child_limit_stack.empty());
        return m_child_limit_stack[level];
      }

      /** The iterator position at the child_limit vector on highest level.
       * \return max_u32 if leaf-node in separated array
       * \throw bit::out_of_range if the iterator is positioned at the
       * virtual root node
       */
      u32 child_limit_index() const
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::Iterator::child_limit_index() called at root node");
        return child_limit_index(length() - 1);
      }

      /** Return the symbol from the current iterator position.
       * \param level = the level from which the symbol is fetched
       * \throw bit::out_of_range if \c level is greater than length() - 1
       */
      u32 symbol(unsigned int level) const
      {
        u32 index = symbol_index(level);
        return m_trie->symbol_array(level).get(index);
      }

      /** Return the symbol from the current iterator position at the
       * highest level.
       * \throw bit::out_of_range if called at virtual root node.
       */
      u32 symbol() const
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::Iterator::symbol() called at root node");
        u32 index = symbol_index(length() - 1);
        return m_trie->symbol_array(length() - 1).get(index);
      }

      /** Return the symbol vector corresponding to the iterator. */
      std::vector<u32> symbol_vec() const
      {
        std::vector<u32> vec(length());
        for (unsigned int l = 0; l < length(); l++)
          vec[l] = symbol(l);
        return vec;
      }

      /** Return the index of the first child after the children of
       * the iterator node at the given level.
       * \param level = the level from which the child_limit is fetched
       * \throw bit::out_of_range if \c level is greater than length() - 1
       * \throw bit::invalid_call if leaf-node in separated array
       */
      u32 child_limit(unsigned int level) const
      {
        u32 index = child_limit_index(level);
        if (index == max_u32)
          throw bit::invalid_call(
            "bit::Trie::child_limit() leaf-node in separated array");

        const Array &array = m_trie->child_limit_array(level);
        if (array.num_elems() == 0)
          return 0;
        if (index >= array.num_elems())
          return array.get(array.num_elems() - 1);
        return m_trie->child_limit_array(level).get(index);
      }

      /** Return the index of the first child after the children of
       * the current node.  Zero MAY be returned if there are no
       * children.
       */
      u32 child_limit() const
      {
        if (is_root())
          return m_trie->num_nodes(0);
        return child_limit(length() - 1);
      }

      /** Return the index of the first child of the iterator node at
       * the given level.  It is safe to call this for the first node
       * of the level; zero is returned.
       * \param level = the level from which the index is fetched.
       * \throw bit::out_of_range if \c level is greater than length() - 1
       * \throw bit::invalid_call if leaf-node in separated array
       */
      u32 first_child(unsigned int level) const
      {
        u32 index = child_limit_index(level);
        if (index == 0)
          return 0;
        if (index == max_u32)
          throw bit::invalid_call(
            "bit::Trie::child_limit() leaf-node in separated array");
        const Array &array = m_trie->child_limit_array(level);
        if (index > array.num_elems())
          return array.get(array.num_elems() - 1);
        return array.get(index - 1);
      }

      /** Return the index of the first child of the current node.  It
       * is safe to call this for the first node of the level; zero is
       * returned.  
       */
      u32 first_child() const
      {
        if (is_root())
          return 0;
        return first_child(length() - 1);
      }

      /** Return the index of the first sibling on the current node at
       * a given level. 
       * \throw bit::out_of_range if level exceeds length() - 1
       */
      u32 first_sibling(unsigned int level) const
      {
        if (level >= length())
          throw bit::out_of_range(
            "bit::Trie::first_sibling(): level out of range");
        if (level == 0)
          return 0;
        return first_child(level - 1);
      }

      /** Return the index of the first sibling on the current
       * node. 
       * \throw bit::out_of_range if called for root
       */
      u32 first_sibling() const
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::first_sibling(): called for root");
        return first_sibling(length() - 1);
      }

      /** Return the index after siblings after the siblings of the
       * current node at a given level. 
       * \throw bit::out_of_range if level exceeds length() - 1
       */
      u32 sibling_limit(unsigned int level) const
      {
        if (level >= length())
          throw bit::out_of_range(
            "bit::Trie::sibling_limit(): level out of range");
        if (level == 0)
          return m_trie->num_nodes(0);
        return child_limit(level - 1);
      }

      /** Return the index after siblings after the siblings of the
       * current node. 
       * \throw bit::out_of_range if called for root
       */
      u32 sibling_limit() const
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::sibling_limit() called for root");
        return sibling_limit(length() - 1);
      }

      /** Is the iterator at the virtual root node? */
      bool is_root() const 
      {
        return length() == 0;
      }

      /** Is the node leaf node in a separated level. */
      bool is_separated_leaf() const
      {
        return (!is_root() && m_child_limit_stack.back() == max_u32);
      }

      /** The number of children the current iterator has at a given
       * level. */
      u32 num_children(unsigned int level) const
      {
        if (is_separated_leaf())
          return 0;

        u32 child_limit = this->child_limit(level);
        if (child_limit == 0)
          return 0;
        return child_limit - first_child(level);
      }

      /** The number of children the current node has. */
      u32 num_children() const
      {
        if (is_root())
          return m_trie->num_nodes(0);
        return num_children(length() - 1);
      }

      //@}



      /** \name Moving the iterator */
      //@{

      /** Iterator positioned at the parent node. 
       * \throw bit::invalid_call if could not move to parent node
       */
      Iterator parent() const
      {
        Iterator it(*this);
        if (!it.goto_parent())
          throw bit::invalid_call("Trie::Iterator::parent() failed");
        return it;
      }

      /** Move the iterator to the virtual root node. */
      void goto_root()
      {
        m_symbol_stack.clear();
        m_child_limit_stack.clear();
      }

      /** Move iterator to the parent of the current node. 
       * \return false if no parent
       */
      bool goto_parent()
      {
        if (is_root())
          return false;
        m_symbol_stack.pop_back();
        m_child_limit_stack.pop_back();
        return true;
      }

      /** Move the iterator to the first child of the current node. 
       * \return false if no children
       */
      bool goto_first_child()
      {
        if (is_separated_leaf())
          return false;

        u32 child_limit = this->child_limit();
        if (child_limit == 0)
          return false;
        u32 first_child = this->first_child();
        assert(first_child <= child_limit);
        if (first_child == child_limit)
          return false;
        
        m_symbol_stack.push_back(first_child);
        m_child_limit_stack.push_back(
          m_trie->compute_non_leaf_index(length() - 1, first_child));
        return true;
      }

      /** Move the iterator to the next sibling.
       * \return false if no more siblings for the current parent 
       * \throw bit::out_of_range if called for root
       */
      bool goto_next_sibling()
      {
        if (is_root())
          throw bit::out_of_range(
            "bit::Trie::Iterator::goto_next_sibling() called for root");
        u32 limit = sibling_limit();
        assert(limit > 0);
        assert(m_symbol_stack.back() < limit);
        if (m_symbol_stack.back() == limit - 1)
          return false;
        m_symbol_stack.back()++;
        m_child_limit_stack.back() =
          m_trie->compute_non_leaf_index(length() - 1, m_symbol_stack.back());
        return true;
      }

      /** Move the iterator to child having the specified symbol. 
       * \param symbol = symbol to find
       * \return false if child with symbol not found
       */
      bool goto_child(u32 symbol)
      {
        if (is_separated_leaf())
          return false;

        u32 limit = child_limit();
        if (limit == 0)
          return false;
        u32 first = first_child();
        assert(first <= limit);
        if (first == limit)
          return false;

        const Array &array = m_trie->symbol_array(length());
        u64 index = binary_search(array, symbol, first, limit);
        if (index == limit)
          return false;
        m_symbol_stack.push_back(index);
        m_child_limit_stack.push_back(
          m_trie->compute_non_leaf_index(length() - 1, index));
        return true;
      }

      /** Move the iterator to the next node in depth-first order. 
       * \return false if no more nodes
       */
      bool goto_next_depth_first()
      {
        if (goto_first_child())
          return true;
        
        while (1) {
          if (goto_next_sibling())
            return true;
          bool ret = goto_parent();
          assert(ret);
          if (is_root())
            return false;
        }
      }

      /** Move the iterator deeper along the first children until a
       * leaf is reached. 
       * \return false if the iterator was not moved at all
       */
      bool goto_leaf()
      {
        bool did_move = false;
        while (goto_first_child())
          did_move = true;
        return did_move;
      }

      /** Move the iterator to the next node in depth-first
       * post-order, i.e., processing children before parents.
       * Assumes that iterator is positioned at root to begin the
       * search.
       *
       * \return false if no more nodes (and the iterator will be at root)
       */
      bool goto_next_depth_first_post()
      {
        if (is_root())
          return goto_leaf();

        if (goto_next_sibling()) {
          goto_leaf();
          return true;
        }

        goto_parent();
        if (is_root())
          return false;
        return true;
      }

      /** Move the iterator depth first to the next node on a given
       * level.
       *
       * \param level = level to iterate (0 = unigram)
       * \return false if no next node (the iterator is left at
       * the original position)
       */
      bool goto_next_on_level(unsigned int level)
      {
	Iterator it(*this);

        while (it.length() > level + 1)
          it.goto_parent();

        if (it.length() == level + 1) {
          if (it.goto_next_sibling())
            goto ok;
          while (1) {
            it.goto_parent();
            if (it.is_root())
              return false;
            if (it.goto_next_sibling())
              break;
          }
        }

	while (it.goto_next_depth_first())
	  if (it.length() == level + 1)
	    goto ok;

	return false;

      ok:
        *this = it;
        return true;
      }

      /** Move the iterator to the first backoff node found in the
       * model.  Backoffing removes symbols from the beginning of the
       * string until the resulting string is found in the model.
       * \throw bit::invalid_call if called at root
       */
      void goto_backoff_full()
      {
        if (is_root())
          throw bit::invalid_call(
            "bit::Trie::Iterator::goto_backoff_full() called at root");

        std::vector<u32> vec(symbol_vec());
        while (!vec.empty()) {
          vec.erase(vec.begin());
          Iterator it(*m_trie);
          bool failed = false;
          for (size_t i = 0; i < vec.size(); i++) {
            if (!it.goto_child(vec[i])) {
              failed = true;
              break;
            }
          }
          if (!failed) {
            *this = it;
            break;
          }
        }
      }

      /** Move the iterator to the backoff node.  Backoffing removes a
       * symbol from the beginning of the string.  If backoff node
       * does not exist, the iterator is left at the original
       * position.
       *
       * \return false if the backoff node does not exist in the model
       * \throw bit::invalid_call if called at root
       */
      bool goto_backoff_once()
      {
        if (is_root())
          throw bit::invalid_call(
            "bit::Trie::Iterator::goto_backoff_once() called at root");

        std::vector<u32> vec(symbol_vec());
        vec.erase(vec.begin());
        Iterator it(*m_trie);
        for (size_t i = 0; i < vec.size(); i++)
          if (!it.goto_child(vec[i]))
            return false;
        *this = it;
        return true;
      }

      //@}



      /** \name Comparing iterators */
      //@{

      /** Test if two iterators point to the same position. */
      bool operator==(const Iterator it)
      {
        return m_symbol_stack == it.m_symbol_stack;
      }

      /** Test if two iterators do not point to the same position. */
      bool operator!=(const Iterator it)
      {
        return m_symbol_stack != it.m_symbol_stack;
      }

      //@}

    private:
      
      /** The trie that the iterator is attached to. */
      const Trie<Array> *m_trie;

      /** The path of symbol vector indices that define the position
       * of the iterator.*/
      std::vector<u32> m_symbol_stack;

      /** The path of child_limit vector indices that define the position
       * of the iterator.  If the last element is max_u32, the current
       * node does not have children. */
      std::vector<u32> m_child_limit_stack;

      friend class Trie;
    
    };

  public:

    /** Reserve space for levels to avoid reallocing and moving arrays. 
     * \param num_levels = number of levels to reserve
     */
    void reserve_levels(unsigned int num_levels)
    {
      m_symbol_arrays.reserve(num_levels);
      m_pointer_arrays.reserve(num_levels);
      m_child_limit_arrays.reserve(num_levels);
    }
  
    /** The number of nodes on a given level.  It is safe to call this
     * for non-existent levels: zero is returned. */
    u32 num_nodes(unsigned int level) const 
    {
      if (level >= m_symbol_arrays.size())
        return 0;
      return m_symbol_arrays.at(level).num_elems();
    }

    /** The bytes allocated for bit buffers of the trie structure.
     * Note that we do not count the stuff required for storing the
     * Array vectors - the cost would be neglible compared to even
     * small size language model.
     */
    u64 size() const {
      u64 ret = 0;
      for (size_t a = 0; a < m_symbol_arrays.size(); a++)
        ret += m_symbol_arrays[a].compressed_size();
      for (size_t a = 0; a < m_child_limit_arrays.size(); a++)
        ret += m_child_limit_arrays[a].compressed_size();
      for (size_t a = 0; a < m_pointer_arrays.size(); a++)
        ret += m_pointer_arrays[a].compressed_size();
      return ret;
    }

    /** Read-only access to symbol arrays. 
     * \param level = the level to access
     */
    const Array &symbol_array(unsigned int level) const
    {
      return m_symbol_arrays.at(level);
    }

    /** Read-only access to child_limit arrays. 
     * \param level = the level to access
     */
    const Array &child_limit_array(unsigned int level) const
    {
      return m_child_limit_arrays.at(level);
    }

    /** Read-only access to pointer arrays. 
     * \param level = the level to access
     */
    const Array &pointer_array(unsigned int level) const
    {
      return m_pointer_arrays.at(level);
    }

    
    /** Insert a new node as a child of the current node or return the
     * iterator if exists alread.  Insertion is should be made only at
     * the end of the child level.  That means that when inserting a
     * 3-gram, it should be greater than the previous 3-gram inserted.
     *
     * \bug Erroneous insertion in the middle is not checked and can
     * produce undefined effects.  At least it would be possible to
     * check that possible previous child is smaller.  Does not check
     * duplicates.
     *
     * \param it = parent node that will get the new child node
     * \param symbol = the symbol of the new node 
     * \return iterator to the inserted node
     */
    Iterator insert(Iterator it, u32 symbol)
    {
      if (it.goto_child(symbol))
        return it;
      return insert_new(it, symbol);
    }

    /** Insert a new string of symbols to the trie of return the
     * iterator if exists already.  Insertion of a string (length n)
     * is allowed only if the prefix (length n-1) exists already.
     *
     * \param vec = symbols of the inserted string 
     * \return iterator positioned at the string.
     *
     * \throw bit::invalid_argument if \c vec is an empty string or
     * the prefix is not found
     */
    template <class T>
    Iterator insert(const std::vector<T> &vec)
    {
      if (vec.empty())
        throw bit::invalid_argument(
          "bit::Trie::insert_new(): trying to add an empty string");
      
      Iterator it(*this);
      for (size_t i = 0; i < vec.size() - 1; i++)
        if (!it.goto_child(vec[i]))
          throw bit::invalid_argument("bit::Trie::insert(): prefix not found");

      if (it.goto_child(vec.back()))
        return it;
      return insert_new(it, vec.back());
    }

    /** Insert a new node as a child of the current node.  Insertion
     * is should be made only at the end of the child level.  That
     * means that when inserting a 3-gram, it should be greater than
     * the previous 3-gram inserted.
     *
     * \bug Erroneous insertion in the middle is not checked and can
     * produce undefined effects.  At least it would be possible to
     * check that possible previous child is smaller.  
     *
     * \param it = parent node that will get the new child node
     * \param symbol = the symbol of the new node 
     * \return iterator to the inserted node
     * \throw bit::invalid_argument if exists already
     */
    Iterator insert_new(Iterator it, u32 symbol)
    {
      unsigned int len = it.length();
      if (len > m_symbol_arrays.size())
        throw bit::invalid_argument(
          "bit::Trie::insert_new(): too long iterator argument");
      
      if (it.goto_child(symbol))
        throw bit::invalid_argument("bit::Trie::insert_new(): duplicate");

      // Update target arrays
      //
      assert(m_symbol_arrays.size() == m_child_limit_arrays.size());
      if (len == m_symbol_arrays.size()) {
        Array empty_array;

        if (m_symbol_arrays.size() == m_symbol_arrays.capacity())
          fprintf(stderr, "DEBUG WARNING: growing array vectors!\n");

        m_symbol_arrays.push_back(empty_array);
        m_child_limit_arrays.push_back(empty_array);
        m_pointer_arrays.push_back(empty_array);
      }
      Array &tgt_symbol_array = m_symbol_arrays.at(len);
      int tgt_symbol_index = tgt_symbol_array.num_elems();
      tgt_symbol_array.set_grow_widen(tgt_symbol_index, symbol);

      // Update child limits on the source level (if not at the first
      // level).
      //
      if (len > 0) {
        Array &src_child_limit_array = m_child_limit_arrays.at(len - 1);
        u32 limit = it.child_limit_index();

        for (u32 i = src_child_limit_array.num_elems(); i < limit; i++)
          src_child_limit_array.set_grow_widen(i, tgt_symbol_index);
        src_child_limit_array.set_grow_widen(limit, tgt_symbol_index + 1);
      }

      // Create the return value
      //
      Iterator tgt_it(it);
      tgt_it.m_symbol_stack.push_back(tgt_symbol_index);
      tgt_it.m_child_limit_stack.push_back(
        compute_non_leaf_index(tgt_it.length() - 1, tgt_symbol_index));

      assert(m_symbol_arrays.size() == m_child_limit_arrays.size());
      assert(m_symbol_arrays.size() == m_pointer_arrays.size());

      return tgt_it;
    }

    /** Insert a new string of symbols to the trie.  Insertion of a
     * string (length n) is allowed only if the prefix (length n-1)
     * exists already.  
     *
     * \param vec = symbols of the inserted string 
     * \return iterator positioned at the new string.
     *
     * \throw bit::invalid_argument if duplicate
     * \throw bit::invalid_argument if \c vec is an empty string or
     * the prefix is not found
     */
    template <class T>
    Iterator insert_new(const std::vector<T> &vec)
    {
      if (vec.empty())
        throw bit::invalid_argument(
          "bit::Trie::insert_new(): trying to add an empty string");
      
      Iterator it(*this);
      for (size_t i = 0; i < vec.size() - 1; i++)
        if (!it.goto_child(vec[i]))
          throw bit::invalid_argument("bit::Trie::insert(): prefix not found");
      return insert_new(it, vec.back());
    }

    /** Find a string of symbols from the trie. 
     * \param vec = symbol string to find
     * \return iterator to the string or empty iterator if not found. 
     * \throw bit::invalid_argument if \c vec is empty
     */
    template <class T>
    Iterator find(const std::vector<T> &vec) const
    {
      if (vec.empty())
        throw bit::invalid_argument(
          "bit::Trie::Iterator::find() called at root");
      Iterator it(*this);
      for (size_t i = 0; i < vec.size(); i++)
        if (!it.goto_child(vec[i]))
          return Iterator();
      return it;
    }

    /** Compress child limit and pointer arrays for certain level.
     * Works only if the Array template type implements
     * recursive_optimal_compress().
     * \param level = level to compress
     * \throw bit::out_of_range if level exceeds number of levels
     */
    void compress(unsigned int level)
    {
      if (level >= m_child_limit_arrays.size())
        throw bit::out_of_range("bit::Trie::compress() level out of bounds");
      m_child_limit_arrays.at(level).recursive_optimal_compress();
      m_pointer_arrays.at(level).recursive_optimal_compress();
    }

    /** Compress all levels. */
    void compress()
    {
      for (size_t l = 0; l < m_child_limit_arrays.size(); l++) {
        m_child_limit_arrays.at(l).recursive_optimal_compress();
        m_pointer_arrays.at(l).recursive_optimal_compress();
      }
    }

    /** Uncompress a level.  It is safe to call this for uncompressed
     * level.
     * \param level = level to uncompress
     * \throw bit::out_of_range if level exceeds number of levels
     */
    void uncompress(unsigned int level)
    {
      if (level >= m_child_limit_arrays.size())
        throw bit::out_of_range("bit::Trie::compress() level out of bounds");
      m_child_limit_arrays.at(level).uncompress();
      m_pointer_arrays.at(level).uncompress();
    }

    /** Uncompress all levels.  It is safe to call this even if some
     * levels are uncompressed. */
    void uncompress()
    {
      for (size_t l = 0; l < m_child_limit_arrays.size(); l++) {
        m_child_limit_arrays.at(l).uncompress();
        m_pointer_arrays.at(l).uncompress();
      }
    }

    /** Check if leafs have been separated for given level. 
     * \param level = level to check
     * \return true if separated
     * \throw bit::out_of_range if \c level is the highest level or
     * greater
     */
    bool is_separated(unsigned int level)
    {
      if (level + 1 >= m_pointer_arrays.size())
        throw bit::out_of_range(
          "bit::Trie::separate_leafs() level out of range");
      return pointer_array(level).num_elems() > 0;
    }

    /** Separate leaf information from non-leaf information (removing
     * possible compression).  This may save space if the level has
     * relatively many leaf nodes and non-leaf nodes have some
     * external information which can be omitted for leaf nodes.
     * Note, that it is not allowed to separate the highest level,
     * because the level does not have children anyway.
     *
     * \warning Invalidates all iterators pointing to the trie on this
     * level or higher.  Lower-lever iterators remain ok.
     *
     * \param level = the level for which the separation is done
     *
     * \throw bit::out_of_range if \c level is the highest level or
     * greater
     *
     * \throw bit::invalid_call if level is already separated
     */
    void separate_leafs(unsigned int level)
    {
      if (level + 1 >= m_pointer_arrays.size())
        throw bit::out_of_range(
          "bit::Trie::separate_leafs() level out of range");
      
      Array &pointer_array = m_pointer_arrays.at(level);
      Array &limit_array = m_child_limit_arrays.at(level);
      Array new_limit_array;

      if (pointer_array.num_elems() != 0)
        throw bit::invalid_call(
          "bit::Trie::separate_leafs() already separated");

      u64 tgt = 0;
      u64 prev_limit = 0;
      for (u64 src = 0; src < limit_array.num_elems(); src++) {
        u32 limit = limit_array.get(src);
        if (limit > prev_limit) {
          new_limit_array.set_grow_widen(tgt, limit);
          tgt++;
        }
        pointer_array.set_grow_widen(src, tgt);
        prev_limit = limit;
      }
      limit_array = new_limit_array;
    }

    /** Undo the leaf separation.  It is safe to call this method even
     * if the leafs are not separated.
     * \warning This call removes possible compression.
     * \param level = level to unseparate
     * \throw bit::out_of_range if level is the highest level or
     * greater
     */
    void unseparate_leafs(unsigned int level)
    {
      if (level + 1 >= m_pointer_arrays.size())
        throw bit::out_of_range(
          "bit::Trie::separate_leafs() level out of range");
      
      CompressedArray &pointer_array = m_pointer_arrays.at(level);
      CompressedArray &limit_array = m_child_limit_arrays.at(level);

      if (pointer_array.num_elems() == 0)
        return;
      
      CompressedArray new_limit_array;
      u32 limit = 0;
      u32 prev_index = 0;
      for (u64 i = 0; i < pointer_array.num_elems(); i++) {
        u32 index = pointer_array.get(i);
        if (index > prev_index) {
          limit = limit_array.get(index - 1);
          prev_index = index;
        }
        new_limit_array.set_grow_widen(i, limit);
      }
      limit_array = new_limit_array;
      pointer_array = CompressedArray();
    }

    /** Display the contents of the child limit arrays. */
    std::string debug_child_limit_arrays_str() const
    {
      std::string ret;
      for (size_t a = 0; a < m_child_limit_arrays.size(); a++) {
        ret.append(str::fmt(256, "===== %d =====\n", a));
        const Array &array = m_child_limit_arrays.at(a);
        ret.append(array.debug_str());
      }
      return ret;
    }

    /** Display the contents of the trie. */
    std::string debug_str() const
    {
      std::string ret;
      for (size_t a = 0; a < m_child_limit_arrays.size(); a++) {
        ret.append(str::fmt(256, "===== %d =====\n", a));
        const Array &symbol_array = m_symbol_arrays.at(a);
        const Array &limit_array = m_child_limit_arrays.at(a);
        const Array &pointer_array = m_pointer_arrays.at(a);
        for (u64 i = 0; i < symbol_array.num_elems(); i++) {
          ret.append(str::fmt(64, "%d:\t%d\t", (int)i, symbol_array.get(i)));
          if (i < pointer_array.num_elems())
            ret.append(str::fmt(64, "%d", pointer_array.get(i)));
          else
            ret.append("-");
          ret.append("\t");
          if (i < limit_array.num_elems())
            ret.append(str::fmt(64, "%d", limit_array.get(i)));
          ret.append("\n");
        }
      }
      return ret;
    }

    /** Write trie in file. 
     * \param file = file stream to write to
     * \throw bit::io_error if write fails
     */
    void write(FILE *file) const
    {
      fputs("TRIE1:", file);
      fprintf(file, "%d:%d:%d:", 
              (int)m_symbol_arrays.size(), 
              (int)m_child_limit_arrays.size(),
              (int)m_pointer_arrays.size());
      if (ferror(file))
        throw bit::io_error(std::string("bit::Trie::write() failed: ") +
                            strerror(errno));
      try {
        for (size_t i = 0; i < m_symbol_arrays.size(); i++)
          m_symbol_arrays[i].write(file);
        for (size_t i = 0; i < m_child_limit_arrays.size(); i++)
          m_child_limit_arrays[i].write(file);
        for (size_t i = 0; i < m_pointer_arrays.size(); i++)
          m_pointer_arrays[i].write(file);
      }
      catch (bit::io_error &e) {
        throw bit::io_error(std::string("bit::Trie::write() failed: ") +
                            e.what());
      }
    }

    /** Read trie from file.  \bug After exception the class may be
     * left in weird state.
     *
     * \param file = file stream to read from
     * \throw bit::io_error if read fails
     */
    void read(FILE *file)
    {
      int symbol_levels;
      int child_limit_levels;
      int pointer_levels;
      int version;
      int ret = fscanf(file, "TRIE%d:%d:%d:%d:", &version,
                       &symbol_levels, &child_limit_levels, &pointer_levels);
      if (ret != 4 || version != 1)
        throw bit::io_error(
          "bit::Trie::read() error while reading header");

      try {
        m_symbol_arrays.resize(symbol_levels);
        m_child_limit_arrays.resize(child_limit_levels);
        m_pointer_arrays.resize(pointer_levels);

        for (size_t a = 0; a < m_symbol_arrays.size(); a++)
          m_symbol_arrays[a].read(file);
        for (size_t a = 0; a < m_child_limit_arrays.size(); a++)
          m_child_limit_arrays[a].read(file);
        for (size_t a = 0; a < m_pointer_arrays.size(); a++)
          m_pointer_arrays[a].read(file);
      }
      catch (bit::io_error &e) {
        throw bit::io_error(
          std::string("bit::CompressedArray::read() read failed") + e.what());
      }
    }

  private:

    /** Compute the non-leaf index corresponding the main index
     * while handling the default and separated mode.
     *
     * \param level = level to access
     * \param symbol_index = the entry for which the non-leaf index
     * is computed.
     * \return max_u32 if the node is leaf, otherwise the non-leaf index
     */
    u32 compute_non_leaf_index(unsigned int level, u32 symbol_index) const
    {
      assert(symbol_index < symbol_array(level).num_elems());
      const Array &pointer_array = this->pointer_array(level);

      // Level is not separated?
      //
      if (pointer_array.num_elems() == 0)
        return symbol_index;

      // Pointer array may be short to omit trailing leaf-nodes.n
      //
      if (symbol_index >= pointer_array.num_elems())
        return max_u32;

      u64 index = pointer_array.get(symbol_index);
      if (index == 0)
        return max_u32;

      u64 prev_index = 0;
      if (symbol_index > 0)
        prev_index = pointer_array.get(symbol_index - 1);
      if (index == prev_index)
        return max_u32;

      return index - 1;
    }


  private:

    /** Arrays containing symbols for each level. */
    std::vector<Array> m_symbol_arrays;

    /** Arrays containing child limits for each level. */
    std::vector<Array> m_child_limit_arrays;

    /** Arrays to store pointers to non-leaf indices.  The node i is
     * non-leaf only if p[i] == p[i-1] + 1.  Otherwise, p[i] ==
     * p[i-1], and the node is a leaf without node-leaf information.
     * The actual non-leaf index for node i is p[i] - 1.
     *
     * In the default mode symbol and child limit arrays have exactly
     * same number of entries, and indices have one-to-one mapping.
     * However, if there are lots of leaf nodes, it may be more
     * efficient to have child limit entires only for non-leaf nodes.
     * Then pointer array is needed for each symbol to indicate the
     * position of possible child limit entry.  Then it is also easy
     * to separate other information for leaf and non-leaf nodes. */
    std::vector<Array> m_pointer_arrays;

  };

};

#endif /* BIT_TRIE_HH */
