#ifndef BIT_BASE_HH
#define BIT_BASE_HH

#include <errno.h>
#include <math.h>
#include <assert.h>
#include <vector>
#include "bit/exceptions.hh"

namespace bit {

  /** Maximum value width in bits. */
  static const unsigned int max_bits_per_value = 32;

  /** Unsigned 32-bit type */
  typedef unsigned int u32;

  /** Maximum value of u64 type. */
  static const u32 max_u32 = (u32)-1;

  /** Signed 32-bit type */
  typedef signed int s32;

  /** Maximum value of s32 type. */
  static const s32 max_s32 = 0x7fffffff;

  /** Unsigned 64-bit type */
  typedef unsigned long long u64;

  /** Signed 64-bit type */
  typedef unsigned long long s64;

  /** Maximum value of u64 type. */
  static const u64 max_u64 = (u64)-1;

  /** Bit-masks containin n lowest bits set to one. */
  static const unsigned int one_masks[] = 
  { 0x00000000, 
    0x00000001, 0x00000003, 0x00000007, 0x0000000f,
    0x0000001f, 0x0000003f, 0x0000007f, 0x000000ff,
    0x000001ff, 0x000003ff, 0x000007ff, 0x00000fff,
    0x00001fff, 0x00003fff, 0x00007fff, 0x0000ffff,
    0x0001ffff, 0x0003ffff, 0x0007ffff, 0x000fffff,
    0x001fffff, 0x003fffff, 0x007fffff, 0x00ffffff,
    0x01ffffff, 0x03ffffff, 0x07ffffff, 0x0fffffff,
    0x1fffffff, 0x3fffffff, 0x7fffffff, 0xffffffff };

  /** Bit-shifts for byte positions. */
  static const int shift[] = { 0, 8, 16, 24, 32 };

  /** The number of 8-bit bytes required to store a number of
   * fixed-width elements.  The bits_per_elem must be 1-32 and the
   * required size must be less than 2^64 bits. */
  inline u64 bytes_required(u64 num_elems, unsigned int bits_per_elem)
  {
    if (bits_per_elem > max_bits_per_value)
      throw bit::invalid_argument(
        "bit::bytes_required(): invalid bits_per_elem");
    return ((u64)num_elems * bits_per_elem + 7) / 8;
  }

  /** The number of highest non-zero bit. */
  inline unsigned int highest_bit(u64 value)
  {
    unsigned int ret = 0;
    while (value != 0) {
      ret++;
      value >>= 1;
    }
    return ret;
  }

  /** Convert float to 32-bit integer representation. */
  inline unsigned int float2uint(float f)
  {
    union {
      float f;
      unsigned int i;
    } t;
    t.f = f;
    return t.i;
  }

  /** Restore float from 32-bit integer representation. */
  inline float uint2float(unsigned int i)
  {
    union {
      float f;
      unsigned int i;
    } t;
    t.i = i;
    return t.f;
  }

  /** Restore a quantized float value using linear quantization
   * centered around zero.  The least-significant bit is used for sign
   * to keep near-zero values with least number of bits.
   *
   * \param i = integer to unquantize
   * \param step = length of the quantization step (0 = no quantization)
   * \return the restored float value
   */
  inline float unquantize_float(u32 i, double step)
  {
    if (step == 0)
      return uint2float(i);
    unsigned int sign = i & 0x1;
    i >>= 1;
    float value = (double)i * step;
    if (sign)
      return -value;
    return value;
  }

  /** Quantize float value using linear quantization centered around
   * zero.  The least-significant bit is used for sign to keep
   * near-zero values with least number of bits.  
   *
   * \param f = float to quantize
   * \param step = length of the quantization step (0 = no quantization)
   * \return the quantized value
   * \throw invalid_argument if quantized value overflows linear scale
   */
  inline u32 quantize_float(float f, double step)
  {
    if (step == 0)
      return float2uint(f);
    assert(step > 0);
    unsigned int sign = 0;
    if (f < 0) {
      sign = 1;
      f = -f;
    }
    double dvalue = (double)f / step;
    if (dvalue > max_s32)
      throw bit::invalid_argument("bit::quantize_float() result too large");
    u32 value = lrint(dvalue);
    return (value << 1) + sign;
  }

};

#endif /* BIT_BASE_HH */
