#ifndef ENDIAN_HH
#define ENDIAN_HH

#include <stdio.h>
#include "assert.h"

/// Tools for handling conversions between different byte orders.
namespace endian {

  /** Returns true if the host is big_endian (used only to initialize
   * bool \c big) */
  inline bool is_big_endian()
  {
    union {
      short i;
      unsigned char buf[2];
    } u;
    u.i = 0xbeaf;
    if (u.buf[0] == 0xbe) {
      assert(u.buf[1] == 0xaf);
      return true;
    }
    else if (u.buf[0] == 0xaf) {
      assert(u.buf[1] == 0xbe);
      return false;
    }
    assert(false);
  }


  /// Is this host big-endian
  static const bool big = is_big_endian();

  /// Write 8-byte value in little-endian format
  template <typename T>
  size_t write8(T v, FILE *file, bool write_big = false)
  {
    assert(sizeof(v) == 8);
    if (big == write_big)
      return fwrite(&v, 8, 1, file);

    char *ptr = (char*)(&v);
    char little[8];
    little[0] = ptr[7];
    little[1] = ptr[6];
    little[2] = ptr[5];
    little[3] = ptr[4];
    little[4] = ptr[3];
    little[5] = ptr[2];
    little[6] = ptr[1];
    little[7] = ptr[0];
    return fwrite(little, 8, 1, file);
  }

  /// Read 8-byte value in little-endian format
  template <typename T>
  size_t read8(T *buf, FILE *file, bool read_big = false)
  {
    if (big == read_big)
      return fread(buf, 8, 1, file);

    char *ptr = (char*)buf;
    char little[8];
    size_t ret = fread(little, 8, 1, file);
    ptr[0] = little[7];
    ptr[1] = little[6];
    ptr[2] = little[5];
    ptr[3] = little[4];
    ptr[4] = little[3];
    ptr[5] = little[2];
    ptr[6] = little[1];
    ptr[7] = little[0];
    return ret;
  }
  
  /// Write 4-byte value in little-endian format
  template <typename T>
  size_t write4(T v, FILE *file, bool write_big = false)
  {
    assert(sizeof(v) == 4);
    if (big == write_big)
      return fwrite(&v, 4, 1, file);

    char *ptr = (char*)(&v);
    char little[4];
    little[0] = ptr[3];
    little[1] = ptr[2];
    little[2] = ptr[1];
    little[3] = ptr[0];
    return fwrite(little, 4, 1, file);
  }

  /// Read 4-byte value in little-endian format
  template <typename T>
  size_t read4(T *buf, FILE *file, bool read_big = false)
  {
    if (big == read_big)
      return fread(buf, 4, 1, file);

    char *ptr = (char*)buf;
    char little[4];
    size_t ret = fread(little, 4, 1, file);
    ptr[0] = little[3];
    ptr[1] = little[2];
    ptr[2] = little[1];
    ptr[3] = little[0];
    return ret;
  }

  /// Write 2-byte value in little-endian format
  template <typename T>
  size_t write2(T v, FILE *file, bool write_big = false)
  {
    assert(sizeof(v) == 2);
    if (big == write_big)
      return fwrite(&v, 2, 1, file);

    char *ptr = (char*)(&v);
    char little[4];
    little[0] = ptr[0];
    little[1] = ptr[1];
    return fwrite(little, 2, 1, file);
  }

  /// Read 2-byte value in little-endian format
  template <typename T>
  size_t read2(T *buf, FILE *file, bool read_big = false)
  {
    if (big == read_big)
      return fread(buf, 2, 1, file);

    char *ptr = (char*)buf;
    char little[4];
    size_t ret = fread(little, 2, 1, file);
    ptr[0] = little[0];
    ptr[1] = little[1];
    return ret;
  }

};

#endif /* ENDIAN_HH */
