#ifndef SIMPLEKNESER_HH
#define SIMPLEKNESER_HH

#include "lm/SymbolMap.hh"
#include "str/str.hh"
#include <map>
#include <vector>
#include "util/Progress.hh"
#include "bit/exceptions.hh"

namespace bit {

  class SimpleKneser {
  public:

    typedef SymbolMap<std::string,int> SymbolMap;
    typedef std::vector<int> Ngram;
    typedef std::vector<std::string> StrNgram;
    typedef std::vector<int> IntVec;
    typedef std::vector<float> FloatVec;

//     struct NgramCompare {
//       bool operator()(const Ngram &a, const Ngram &b) const
//       {
//         if (a.size() < b.size())
//           return true;
//         if (a.size() > b.size())
//           return false;
//         return a < b;
//       }
//     };

    // d2 measure needs float nominator and int denominator.  For
    // pruned ngrams the integer denominator is zero.
    struct FloatInt {
      FloatInt() : f(0), i(0) { }
      FloatInt(float f, int i) : f(f), i(i) { }
      float f;
      int i;

      void add(const FloatInt &value)
      {
        f += value.f;
        i += value.i;
        assert(i > 0);
      }

      void sub(const FloatInt &value)
      {
        f -= value.f;
        i -= value.i;
        assert(i > 0);
      }
      
      float value() const
      {
        return f / i;
      }

    };

    // Maps used for caching statistics 
    typedef std::map<Ngram, int> IntMap;
    typedef std::map<Ngram, float> FloatMap;
    typedef std::map<Ngram, FloatInt> FloatIntMap;

    SimpleKneser()
    {
      init();
    }

    SimpleKneser(const std::string &str)
    {
      init();
      StrNgram str_ngram = str::split(str, " \t", true);
      for (size_t i = 0; i < str_ngram.size(); i++)
        m_symbol_map.insert(str_ngram[i]);
      m_sentence_start_id = m_symbol_map.index(m_sentence_start_str);
      m_num_events = m_symbol_map.size() - 1;
    }

    void init()
    {
      m_progress_skip = 7351;
      m_sentence_start_str = "<s>";
      m_sentence_start_id = -1;
    }

    void set_discounts(const std::string &str)
    {
      m_discounts = str::float_vec(str);
      if (m_discounts.empty())
        throw bit::invalid_argument(
          "bit::SimpleKneser::set_discounts() empty vector");
      // Dummy value to get unigram discount on index 1
      m_discounts.insert(m_discounts.begin(), -1e30);

      m_beta_discounts = m_discounts;
      for (size_t i = m_beta_discounts.size() - 1; i > 0; i--)
        m_beta_discounts[i-1] *= m_beta_discounts[i];
    }

    float get_discount(unsigned int order) const
    {
      if (m_discounts.empty())
        throw bit::invalid_call(
          "bit::SimpleKneser::get_discount() discounts not set");
      if (order == 0)
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_discount() trying to get zero order");
      if (order >= m_discounts.size())
        return m_discounts.back();
      return m_discounts.at(order);
    }

    float get_beta_discount(unsigned int order) const
    {
      if (m_beta_discounts.empty())
        throw bit::invalid_call(
          "bit::SimpleKneser::get_beta_discount() beta_discounts not set");
      if (order == 0)
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_beta_discount() trying to get zero order");
      if (order >= m_beta_discounts.size())
        return m_beta_discounts.back();
      return m_beta_discounts.at(order);
    }

    Ngram ngram(const std::string &str) const
    {
      StrNgram str_ngram = str::split(str, " \t", true);
      Ngram ngram(str_ngram.size());
      for (size_t i = 0; i < str_ngram.size(); i++)
        ngram[i] = m_symbol_map.index(str_ngram[i]);
      return ngram;
    }

    int get_count(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_counts.find(ngram);
      if (it == m_counts.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_count() ngram not found");
      return it->second;
    }

    int get_sum_nonzero_xg(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_sum_nonzero_xg.find(ngram);
      if (it == m_sum_nonzero_xg.end())
        return -1;
      return it->second;
    }

    int get_sum_nonzero_xgx(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_sum_nonzero_xgx.find(ngram);
      if (it == m_sum_nonzero_xgx.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_sum_nonzero_xgx() ngram not found");
      return it->second;
    }

    int get_sum_nonzero_gx(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_sum_nonzero_gx.find(ngram);
      if (it == m_sum_nonzero_gx.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_sum_nonzero_gx() ngram not found");
      assert(it->first.empty() || it->first.back() != m_sentence_start_id);
      return it->second;
    }

    int get_sum_xg_not_pruned(const Ngram &ngram) const
    {
      if (!ngram.empty() && ngram.back() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_sum_xg_not_pruned() sentence start");
      IntMap::const_iterator it = m_sum_xg_not_pruned.find(ngram);
      if (it == m_sum_xg_not_pruned.end())
        return 0;
      return it->second;
    }

    int get_sum_nonzero_xg_not_pruned(const Ngram &ngram) const
    {
      if (!ngram.empty() && ngram.back() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_sum_nonzero_xg_not_pruned() sentence start");
      IntMap::const_iterator it = m_sum_nonzero_xg_not_pruned.find(ngram);
      if (it == m_sum_nonzero_xg_not_pruned.end())
        return 0;
      return it->second;
    }

    int get_sum_nonzero_gx_not_pruned(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_sum_nonzero_gx_not_pruned.find(ngram);
      if (it == m_sum_nonzero_gx_not_pruned.end())
        return 0;
      return it->second;
    }

    float get_d1(const Ngram &ngram) const
    {
      FloatMap::const_iterator it = m_d1.find(ngram);
      if (it == m_d1.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_d1() ngram not found");
      assert(it->first.empty() || it->first.back() != m_sentence_start_id);
      return it->second;
    }

    float get_d2(const Ngram &ngram) const
    {
      FloatInt value = get_d2_pair(ngram);
      return value.f / value.i;
    }

    const FloatInt &get_d2_pair(const Ngram &ngram) const
    {
      FloatIntMap::const_iterator it = m_d2.find(ngram);
      if (it == m_d2.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_d2_pair() ngram not found");
      assert(it->first.empty() || it->first.back() != m_sentence_start_id);
      return it->second;
    }

    FloatInt &get_d2_pair(const Ngram &ngram) 
    {
      FloatIntMap::iterator it = m_d2.find(ngram);
      if (it == m_d2.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_d2_pair() ngram not found");
      assert(it->first.empty() || it->first.back() != m_sentence_start_id);
      return it->second;
    }

    float get_beta_numerator(const Ngram &ngram) const
    {
      if (is_pruned(ngram))
        return 0;
      return get_count(ngram) - get_sum_xg_not_pruned(ngram) +
        get_beta_discount(ngram.size() + 1) * 
        get_sum_nonzero_xg_not_pruned(ngram);
    }

    float get_beta_denominator(const Ngram &ngram) const
    {
      FloatMap::const_iterator it = m_beta_denominator.find(ngram);
      if (it == m_beta_denominator.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::get_beta_denominator() ngram not found");
      return it->second;
    }

    void add_count(const Ngram &ngram, int count)
    {
      m_counts[ngram] += count;
    }

    void read_counts(FILE *file)
    {
      std::string line;
      StrNgram str_ngram;
      Progress p(m_progress_skip);
      p.set_report_string("reading counts:");
      while (str::read_line(line, file, true)) {
        str_ngram = str::split(line, " \t", true);
        if (str_ngram.size() < 2)
          throw bit::io_error(
            "bit::SimpleKneser::read_counts(): invalid line");
        int count = str::str2long(str_ngram.back());
        str_ngram.pop_back();

        Ngram ngram(str_ngram.size());
        for (size_t i = 0; i < str_ngram.size(); i++)
          ngram[i] = m_symbol_map.insert(str_ngram[i]);

        m_counts[ngram] += count;
        p.step();
      }
      p.finish();

      assert(m_sentence_start_id < 0);
      m_sentence_start_id = m_symbol_map.index(m_sentence_start_str);
      m_num_events = m_symbol_map.size() - 1;
    }

    void compute_modified_counts()
    {
      assert(!m_counts.empty());
      Progress p(m_progress_skip, m_counts.size() * 3);
      p.set_report_string("mod counts:");

      for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++)
      {
        p.step();
        if (it->first.back() == m_sentence_start_id)
          continue;
        if (it->first.size() > 1) {
          m_sum_nonzero_xg[backoff(it->first)]++;
          m_sum_xg_not_pruned[backoff(it->first)] += it->second;
          m_sum_nonzero_xg_not_pruned[backoff(it->first)]++;
        }
        m_sum_nonzero_gx[parent(it->first)]++;
      }

      for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++)
      {
        p.step();
        if (m_sum_nonzero_xg[it->first] == 0)
          m_sum_nonzero_xg[it->first] = it->second;
      }

      for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++)
      {
        p.step();
        if (it->first.back() == m_sentence_start_id)
          continue;
        m_sum_nonzero_xgx[parent(it->first)] += get_sum_nonzero_xg(it->first);
      }

      p.finish();

      m_sum_nonzero_gx_not_pruned = m_sum_nonzero_gx;
    }

    float ngram_prob(Ngram ngram) const
    {
      double ret = 1;
      while (!ngram.empty()) {
        if (ngram.size() == 1 && ngram.back() == m_sentence_start_id)
          break;
        ret *= prob_kn(ngram);
        ngram.pop_back();
      }
      return ret;
    }

    void compute_d1()
    {
      double min_d1 = 1e100;
      double max_d1 = -1e100;

      assert(!m_sum_nonzero_xg.empty());
      Progress p(m_progress_skip, m_counts.size());
      p.set_report_string("d1:");

      for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++)
      {
        p.step();
        if (it->first.back() == m_sentence_start_id)
          continue;
        double orig = prob_kn(it->first);
        double lower = prob_kn_lower(it->first);

        double d1 = ngram_prob(it->first) * log10(orig / lower);
        m_d1[it->first] = d1;

        if (d1 < min_d1)
          min_d1 = d1;
        if (d1 > max_d1)
          max_d1 = d1;

        if (0) {
          fprintf(stderr, "d1 = %12g  %s\n", m_d1[it->first],
                  ngram_str(it->first).c_str());
        }
        
      }
      p.finish();

      fprintf(stderr, "min_d1 = %g\n", min_d1);
      fprintf(stderr, "max_d1 = %g\n", max_d1);
    }

    void compute_d2()
    {
      assert(!m_d1.empty());
      Progress p(m_progress_skip, m_counts.size());
      p.set_report_string("d2:");

      FloatMap::iterator it = m_d1.end(); 
      while (it != m_d1.begin()) {
        it--;
        p.step();

        FloatInt value(it->second, 1);
        FloatInt &d2 = m_d2[it->first];

        d2.add(value);
        if (it->first.size() > 1)
          m_d2[parent(it->first)].add(d2);

        if (0) {
          fprintf(stderr, "%12g / %d %s\n", d2.f, d2.i, 
                  ngram_str(it->first).c_str());
        }
      }
      p.finish();
    }

    void compute_beta_denominator()
    {
      fprintf(stderr, "FATAL: compute_beta_denominator(): "
              "denominator is computed incorrecly\n");
      abort();

      assert(!m_sum_xg_not_pruned.empty());
      Progress p(m_progress_skip, m_counts.size());
      p.set_report_string("beta denominator:");
      m_beta_denominator.clear();

      for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++)
      {
        p.step();
        if (it->first.back() == m_sentence_start_id)
          continue;

        double numerator = get_beta_numerator(it->first);
        m_beta_denominator[parent(it->first)] += numerator;
      }
      p.finish();
    }

    bool is_pruned(const Ngram &ngram) const
    {
      if (ngram.size() == 1)
        return false;
      return get_d2_pair(ngram).i == 0;
    }

    /** Prune ngram (and possible children), and modify d2 measure for
     * the parent ngrams.  The pruned ngrams are marked by setting the
     * integer denominator of d2 to zero. 
     *
     * \throw bit::invalid_argument if ngram already pruned, or ngram
     * is 1-gram, or ngram not found
     */
    void prune_ngram(Ngram ngram)
    {
      assert(!m_d2.empty());

      if (ngram.size() < 2)
        throw bit::invalid_argument(
          "bit::SimpleKneser::prune_ngram() ngram order < 2");

      FloatIntMap::iterator it = m_d2.find(ngram);
      if (it == m_d2.end())
        throw bit::invalid_argument(
          "bit::SimpleKneser::prune_ngram() ngram not found");
      
      if (it->second.i == 0)
        throw bit::invalid_argument(
          "bit::SimpleKneser::prune_ngram() ngram already pruned");

      while (ngram.size() > 1) {
        ngram.pop_back();
        if (ngram.size() == 1 && ngram.back() == m_sentence_start_id)
          break;
        get_d2_pair(ngram).sub(it->second);
      }

      // Mark ngram and its children pruned
      //
      fprintf(stderr, "pruned %12g %12g / %d %s\n", 
              get_d1(it->first), it->second.f, it->second.i,
              ngram_str(it->first).c_str());

      it->second.i = 0;
      m_sum_xg_not_pruned[backoff(it->first)] -= get_count(it->first);
      m_sum_nonzero_xg_not_pruned[backoff(it->first)]--;
      m_sum_nonzero_gx_not_pruned[parent(it->first)]--;
      size_t len = it->first.size();
      while (1) {
        it++;
        if (it == m_d2.end())
          break;
        if (it->first.size() <= len)
          break;
        if (it->second.i > 0) {
          fprintf(stderr, "pruned %12g %12g / %d %s\n", 
                  get_d1(it->first),
                  it->second.f, it->second.i,
                  ngram_str(it->first).c_str());
          it->second.i = 0;
          m_sum_xg_not_pruned[backoff(it->first)] -= get_count(it->first);
          m_sum_nonzero_xg_not_pruned[backoff(it->first)]--;
          m_sum_nonzero_gx_not_pruned[parent(it->first)]--;
        }
      }
    }

    void prune(float threshold)
    {
      FloatIntMap::iterator it = m_d2.end();
      while (it != m_d2.begin()) {
        it--;
        if (it->second.value() < threshold)
          prune_ngram(it->first);
      }
    }

    Ngram parent(const Ngram &ngram) const
    {
      assert(!ngram.empty());
      return Ngram(ngram.begin(), ngram.end() - 1);
    }

    Ngram backoff(const Ngram &ngram) const
    {
      assert(ngram.size() > 1);
      return Ngram(ngram.begin() + 1, ngram.end());
    }

    float inter_kn(const Ngram &ngram) const
    {
      IntMap::const_iterator it = m_sum_nonzero_gx.find(ngram);
      if (it == m_sum_nonzero_gx.end())
        return 1;
      return it->second * get_discount(ngram.size() + 1) / 
        get_sum_nonzero_xgx(ngram);
    }

    float prob_kn_lower(Ngram ngram) const
    {
      if (ngram.empty())
        throw bit::invalid_argument(
          "bit::SimpleKneser::prob_kn_lower() empty ngram");

      if (ngram.back() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::SimpleKneser::prob_kn_lower() sentence start");

      double scale = 1;
      double ret = 0;
      while (1) {
        scale *= inter_kn(parent(ngram));
        ngram.erase(ngram.begin());
        if (ngram.empty())
          break;

        int xg = get_sum_nonzero_xg(ngram);
        assert(xg != 0);
        if (xg >= 0) {
          double numerator = xg - get_discount(ngram.size());
          double denominator = get_sum_nonzero_xgx(parent(ngram));
          ret += scale * numerator / denominator;
        }
      }
      ret += scale / m_num_events;
      return ret;
    }

    float prob_kn(const Ngram &ngram) const
    {
      double numerator = 0;
      double denominator = 1;
      int xg = get_sum_nonzero_xg(ngram);
      assert(xg != 0);
      if (xg > 0) {
        numerator = xg - get_discount(ngram.size());
        denominator = get_sum_nonzero_xgx(parent(ngram));
      }
      return numerator / denominator + prob_kn_lower(ngram);
    }

    float inter_beta(const Ngram &ngram) const
    {
      assert(!m_beta_denominator.empty());
      assert(!m_sum_nonzero_xg_not_pruned.empty());

      int value = get_sum_nonzero_gx_not_pruned(ngram);
      if (value == 0)
        return 1;
      return value * get_beta_discount(ngram.size() + 1) / 
         get_beta_denominator(ngram);
    }

    float prob_beta_lower(Ngram ngram) const
    {
      if (ngram.empty())
        throw bit::invalid_argument(
          "bit::SimpleKneser::prob_beta_lower() empty ngram");

      if (ngram.back() == m_sentence_start_id)
        throw bit::invalid_argument(
          "bit::SimpleKneser::prob_beta_lower() sentence start");

      double scale = 1;
      double ret = 0;
      while (1) {
        scale *= inter_beta(parent(ngram));
        ngram.erase(ngram.begin());
        if (ngram.empty())
          break;

        if (is_pruned(ngram))
          continue;

        double numerator = get_beta_numerator(ngram) 
          - get_beta_discount(ngram.size());
        assert(numerator >= 0);
        double denominator = get_beta_denominator(parent(ngram));
        ret += scale * numerator / denominator;
      }
      ret += scale / m_num_events;
      return ret;
    }

    float prob_beta(const Ngram &ngram) const
    {
      double numerator = 0;
      double denominator = 1;

      if (!is_pruned(ngram)) {
        numerator = get_beta_numerator(ngram) - 
          get_beta_discount(ngram.size());
        denominator = get_beta_denominator(parent(ngram));
      }
      return numerator / denominator + prob_beta_lower(ngram);
    }

    std::string ngram_str(const Ngram &ngram)
    {
      std::string ret;
      for (size_t i = 0; i < ngram.size(); i++) {
        if (i > 0)
          ret.append(" ");
        ret.append(m_symbol_map.at(ngram[i]));
      }
      return ret;
    }

    void write_beta_arpa(FILE *file)
    {
      unsigned int max_order = 0;
      for (size_t o = 1;; o++) {
      
        if (max_order > 0 && o > max_order)
          break;

        fprintf(stderr, "writing %zd-grams (max order %d)\n", o, max_order);
        fprintf(file, "\n\\%zd-grams:\n", o);

        for (IntMap::iterator it = m_counts.begin(); it != m_counts.end(); it++) 
        {
          if (is_pruned(it->first))
            continue;

          if (it->first.size() > max_order)
            max_order = it->first.size();

          if (it->first.size() != o)
            continue;

          float prob = -99;
          if (it->first.back() != m_sentence_start_id)
            prob = log10(prob_beta(it->first));

          fprintf(file, "%g\t%s", prob,
                  ngram_str(it->first).c_str());
          if (get_sum_nonzero_gx_not_pruned(it->first) > 0)
            fprintf(file, "\t%g", log10(inter_beta(it->first)));
          fputs("\n", file);
        }
      }
    }

  private:
    std::string m_sentence_start_str;
    int m_sentence_start_id;

    int m_num_events;
    int m_progress_skip;
    SymbolMap m_symbol_map;
    FloatVec m_discounts;
    FloatVec m_beta_discounts;
    IntMap m_counts;
    IntMap m_sum_nonzero_xg;
    IntMap m_sum_nonzero_xgx;
    IntMap m_sum_nonzero_gx;
    IntMap m_sum_xg_not_pruned;
    IntMap m_sum_nonzero_xg_not_pruned;
    IntMap m_sum_nonzero_gx_not_pruned;
    FloatMap m_beta_denominator;
    FloatMap m_d1;
    FloatIntMap m_d2;
  };

};

#endif /* SIMPLEKNESER_HH */
