#include <assert.h>
#include "str/str.hh"
#include "lm/ArpaReader.hh"
#include "bit/exceptions.hh"

namespace bit {

  ArpaReader::ArpaReader(FILE *file)
  {
    opt.throw_header_mismatch = true;
    opt.show_progress = false;
    opt.ignore_symbols.push_back("<UNK>");
    reset(file);
  }

  void 
  ArpaReader::reset(FILE *file)
  {
    header.order = 0;
    header.num_ngrams_total = 0;
    header.num_ngrams.clear();
    m_current_order = -1;
    m_ngrams_read = 0;
    m_file = file;
    end_reached = false;
  }

  void
  ArpaReader::read_header()
  {
    try {
      std::string line;
      std::vector<std::string> fields;

      // Read until \data\ reached
      //
      while (1) {
        if (!str::read_line(line, m_file, true))
          throw io_error(
            "lm::ArpaReader::read_header(): keyword \\data\\ not found");
        if (line == "\\data\\")
          break;
      }

      // Parse order counts
      //
      while (1) {
        if (!str::read_line(line, m_file, true))
          throw io_error("lm::ArpaReader::read_header(): unexpected eof");
        str::clean(line, " \t");
        if (line.empty())
          continue;

        if (line == "\\1-grams:") {
          m_current_order = 0;
          m_ngrams_read = 0;
          break;
        }

        if (line.substr(0, 6) != "ngram ")
          throw line;
        fields = str::split(line.substr(6), "=", false, 2);
        if (fields.size() != 2)
          throw line;

        try {
          int o = str::str2long(fields[0]);
          int c = str::str2long(fields[1]);
          if (o != header.order + 1 || c < 0)
            throw line;

          header.num_ngrams.push_back(c);
          header.order++;
          header.num_ngrams_total += c;
        }
        catch (std::exception &e) {
          throw line;
        }
      }
    }
    catch (std::string &str) {
      throw io_error(
        std::string("lm::ArpaReader::read_header(): invalid line: ") + str);
    }
  }

  bool
  ArpaReader::read_ngram()
  {
    try {
      if (m_current_order < 0)
        throw std::string("called before read_header()");

      if (end_reached)
        return false;

      std::string line;
      std::vector<std::string> fields;
      
    restart:
      while (1) {
        if (!str::read_line(line, m_file, true)) {
          if (opt.throw_header_mismatch)
            throw std::string("end of file before \\end\\\n");
          fprintf(stderr, "WARNING: end of file before \\end\\\n");
          line = "\\end\\";
        }
        str::clean(line, " \t");
        if (line.empty())
          continue;

        if (line[0] == '\\') {

          if (m_ngrams_read != header.num_ngrams.at(m_current_order)) {
            if (opt.throw_header_mismatch)
              throw str::fmt(256, 
                             "expected number of %d-grams was %d, got %d\n",
                             m_current_order + 1, 
                             header.num_ngrams[m_current_order], 
                             m_ngrams_read);
            fprintf(stderr, "WARNING: expected number of %d-grams was %d, "
                    "got %d\n", m_current_order + 1, 
                    header.num_ngrams[m_current_order], m_ngrams_read);
            header.num_ngrams[m_current_order] = m_ngrams_read;
          }

          if (line == "\\end\\") {
            if (m_current_order < header.order - 1) {
              if (opt.throw_header_mismatch)
                throw str::fmt(256, "\\end\\ after %d-grams, expected up to "
                               "%d-grams\n", m_current_order + 1, 
                               header.order);
              fprintf(stderr, "WARNING: \\end\\ after %d-grams, "
                      "expected up to %d-grams\n", m_current_order + 1, 
                      header.order);
            }
            end_reached = true;
            return false;
          }

          int order;
          int ret = sscanf(line.c_str(), "\\%d-grams:", &order);
          if (ret != 1)
            throw std::string("invalid keyword on line: " + line);

          assert(m_ngrams_read <= header.num_ngrams[m_current_order]);
          if (m_ngrams_read < header.num_ngrams[m_current_order]) {
            if (opt.throw_header_mismatch)
              throw str::fmt(256, "got %d n-grams for order %d, %d expected\n",
                             m_ngrams_read, m_current_order + 1, 
                             header.num_ngrams[m_current_order]);
            fprintf(stderr, 
                    "WARNING: got %d n-grams for order %d, %d expected\n",
                    m_ngrams_read, m_current_order + 1, 
                    header.num_ngrams[m_current_order]);
            header.num_ngrams[m_current_order] = m_ngrams_read;
          }

          if (order != m_current_order + 2)
            throw std::string("invalid order on line: " + line);

          m_ngrams_read = 0;
          m_current_order++;

          assert(m_current_order <= header.order);
          if (m_current_order == header.order) {
            if (opt.throw_header_mismatch)
              throw str::fmt(256, "did not expect %d-grams\n", m_current_order+1);
            fprintf(stderr, "WARNING: did not expect %d-grams\n", 
                    m_current_order + 1);
            header.num_ngrams.push_back(0);
            header.order++;
          }

          return false;
        }
        else
          break;
      }

      // Parse fields of the ngram line
      //
      fields = str::split(line, " \t", true);
      if ((int)fields.size() < m_current_order + 2)
        throw std::string("too few fields on line: " + line);
      if ((int)fields.size() > m_current_order + 3)
        throw std::string("too many fields on line: " + line);
      if ((int)fields.size() == m_current_order + 3) {
        try {
          ngram.backoff = str::str2float(fields[m_current_order + 2]);
        }
        catch (std::exception &e) {
          throw std::string("invalid backoff value on line: ") + line;
        }
      }
      else
        ngram.backoff = 0;

      try {
        ngram.log_prob = str::str2float(fields[0]);
      }
      catch (std::exception &e) {
        throw std::string("invalid log-probability on line: ") + line;
      }

      ngram.symbols.clear();

      for (int i = 1; i < m_current_order + 2; i++)
        if (std::find(opt.ignore_symbols.begin(), opt.ignore_symbols.end(), 
                      fields[i]) != opt.ignore_symbols.end())
        {
          fputs("WARNING: ignored ngram:", stderr);
          for (int i = 1; i < m_current_order + 2; i++)
            fprintf(stderr, " %s", fields[i].c_str());
          fputs("\n", stderr);
          header.num_ngrams[m_current_order]--;
          goto restart;
        }

      if (m_current_order == 0)
        ngram.symbols.push_back(symbol_map.insert_new(fields[1]));
      else {
        try {
          for (int i = 1; i < m_current_order + 2; i++)
            ngram.symbols.push_back(symbol_map.index(fields[i]));
        }
        catch (std::string &str) {
          throw str::fmt(256, "invalid symbol in %d-gram: ", 
                         m_current_order + 1) + line;
        }
      }

      m_ngrams_read++;
    }
    catch (std::string &str) {
      throw io_error(std::string("lm::ArpaReader::read_ngram(): ") + str);
    }

    return true;
  }

  struct Compare {
    Compare(const std::vector<ArpaReader::Ngram> &ngrams)
      : ngrams(ngrams) { }
    bool operator()(int a, int b) const 
    {
      return ngrams[a] < ngrams[b];
    }
    const std::vector<ArpaReader::Ngram> &ngrams;
  };

  bool
  ArpaReader::read_order_ngrams(bool sort)
  {
    if (end_reached)
      return false;
  
    if (opt.show_progress)
      fprintf(stderr, "reading %d-grams\n", m_current_order + 1);
    order_ngrams.clear();
    sorted_order.clear();
    while (read_ngram())
      order_ngrams.push_back(ngram);
    if (opt.show_progress)
      fprintf(stderr, "got %d\n", (int)order_ngrams.size());

    if (sort) {
      if (opt.show_progress)
        fprintf(stderr, "sorting\n");
      sorted_order.resize(order_ngrams.size());
      for (int i = 0; i < (int)order_ngrams.size(); i++)
        sorted_order[i] = i;
      std::sort(sorted_order.begin(), sorted_order.end(), 
                Compare(order_ngrams));
    }

    return true;
  }

};
