#ifndef TIMER_HH
#define TIMER_HH

#include <unistd.h>
#include <time.h>
#include <sys/times.h>
#include <stdio.h>

/** A class for measuring CPU time in seconds. */
class Timer {
public:
  /** Default constructor. */
  Timer();

  /** Starts the clock without resetting it. */
  void start();

  /** Stops the clock (can be restarted) */
  void stop();

  /** Reset the clock. */
  void reset();

  /** Return the duration in ticks. */
  clock_t user_ticks() const;
  clock_t sys_ticks() const;
  clock_t total_ticks() const;

  /** Return the duration in seconds. */
  float user_sec() const;
  float sys_sec() const;
  float total_sec() const;

protected:
  clock_t m_user_ticks;
  clock_t m_sys_ticks;
  clock_t m_user_start;
  clock_t m_sys_start;
  long m_ticks_per_sec;
  bool m_running;
};

inline
Timer::Timer()
  : m_user_ticks(0), m_sys_ticks(0)
{
  m_running = false;
  m_ticks_per_sec = sysconf(_SC_CLK_TCK);
  if (m_ticks_per_sec <= 0) {
    perror("ERROR: sysconf(_SC_CLK_TCK) failed");
    exit(1);
  }
}

inline
void
Timer::start()
{
  assert(!m_running);
  m_running = true;

  struct tms tms;
  clock_t ret = times(&tms);
  if (ret < 0) {
    perror("ERROR: times() failed");
    exit(1);
  }
  m_user_start = tms.tms_utime;
  m_sys_start = tms.tms_stime;
}

inline
void
Timer::stop()
{
  assert(m_running);
  m_running = false;

  struct tms tms;
  clock_t ret = times(&tms);
  if (ret < 0) {
    perror("ERROR: times() failed");
    exit(1);
  }
  m_user_ticks += tms.tms_utime - m_user_start;
  m_sys_ticks += tms.tms_stime - m_sys_start;
}

inline
void
Timer::reset()
{
  assert(!m_running);
  m_user_ticks = 0;
  m_sys_ticks = 0;
}

inline
clock_t
Timer::user_ticks() const
{
  assert(!m_running);
  return m_user_ticks;
}

inline
clock_t
Timer::sys_ticks() const
{
  assert(!m_running);
  return m_sys_ticks;
}

inline
clock_t
Timer::total_ticks() const
{
  assert(!m_running);
  return m_user_ticks + m_sys_ticks;
}

inline
float
Timer::user_sec() const
{
  assert(!m_running);
  return (float)m_user_ticks / (float)m_ticks_per_sec;
}

inline
float
Timer::sys_sec() const
{
  assert(!m_running);
  return (float)m_sys_ticks / (float)m_ticks_per_sec;
}

inline
float
Timer::total_sec() const
{
  assert(!m_running);
  return total_ticks() / (float)m_ticks_per_sec;
}

#endif /* TIMER_HH */
