#include "conf/conf.hh"
#include "io/io.hh"
#include "str/str.hh"
#include "lm/LM.hh"

using namespace bit;

conf::Config config;
LM lm;

int
main(int argc, char *argv[])
{
  try {
    config("usage: pack [OPTION...]\n")
      ('h', "help", "", "", "display help")
      ('\0', "arpa=FILE", "arg", "", "read ARPA language model")
      ('\0', "blm=FILE", "arg", "", "read binary blm model")
      ('\0', "out-arpa", "arg", "", "write ARPA model")
      ('\0', "out-blm", "arg", "", "write blm model")
      ('\0', "compress", "", "", "compress blm model before writing")
      ('\0', "separate", "", "", "separate tables if compression achieved")
      ('\0', "float-bits=INT", "arg", "", 
       "linear quantization of floats using INT bits")
      ;
    config.default_parse(argc, argv);
    if (config.arguments.size() != 0)
      config.print_help(stderr, 1);

    // Read the language model
    //
    if (config["arpa"].specified) {
      if (config["blm"].specified) {
        fprintf(stderr, "options --arpa and --blm not allowed together\n");
        exit(1);
      }
      lm.read_arpa(io::Stream(config["arpa"].get_str()), "<s>", "</s>", true);
    }
    else if (config["blm"].specified) {
      lm.read(io::Stream(config["blm"].get_str()));
    }
    else {
      fprintf(stderr, "option --arpa or --blm required\n");
      exit(1);
    }
    fprintf(stderr, "model order %d\n", lm.order());
    fprintf(stderr, "model size %lld\n", lm.size());

    // Quantization
    //
    if (config["float-bits"].specified) {
      int bits = config["float-bits"].get_int();
      if (bits <= 0 || bits >= 32) {
	fprintf(stderr, "invalid float-bits specified: %s\n",
		config["float-bits"].get_c_str());
	exit(1);
      }

      fprintf(stderr, "quantizing floats linearly in %d bits\n", bits);
      lm.linear_quantization(bits);
    }

    // Separation and compressing
    //

    bool compress = config["compress"].specified;

    if (config["separate"].specified) {
      if (compress)
        lm.compress_trie();
      for (unsigned int l = 0; l < lm.order() - 1; l++) {
        u64 symbol_size = lm.symbol_array(l).compressed_size();
        u64 pointer_size = lm.pointer_array(l).compressed_size();
        u64 child_limit_size = lm.child_limit_array(l).compressed_size();
        u64 score_size = lm.score_array(l).data_len();
        u64 backoff_size = lm.backoff_array(l).data_len();

        fprintf(stderr, "separating order %d %s\n", l + 1,
                compress ? "with compression" : "without compression");
        u64 orig_size = lm.size();
        fprintf(stderr, "original size: %lld\n", orig_size);
        lm.separate_leafs(l);
        if (compress)
          lm.compress_trie(l);
        u64 new_size = lm.size();
        fprintf(stderr, "new size:      %lld (%.2f%%)\n", new_size,
                100.0*new_size / orig_size);
        if (new_size > orig_size) {
          lm.unseparate_leafs(l);
          if (compress)
            lm.compress_trie(l);

          u64 symbol_size2 = lm.symbol_array(l).compressed_size();
          u64 pointer_size2 = lm.pointer_array(l).compressed_size();
          u64 child_limit_size2 = lm.child_limit_array(l).compressed_size();
          u64 score_size2 = lm.score_array(l).data_len();
          u64 backoff_size2 = lm.backoff_array(l).data_len();
          assert(symbol_size == symbol_size2);
          assert(pointer_size == pointer_size2);
          assert(child_limit_size == child_limit_size2);
          assert(score_size == score_size2);
          assert(backoff_size == backoff_size2);
        }
      }
    }

    else if (compress) {
      lm.compress_trie();
      fprintf(stderr, "model size %lld after compression\n", lm.size());
    }
    
    // Write models
    //
    if (config["out-arpa"].specified) {
      fprintf(stderr, "writing ARPA model: %s\n", 
              config["out-arpa"].get_c_str());
      lm.write_arpa(io::Stream(config["out-arpa"].get_str(), "w"));
    }
    if (config["out-blm"].specified) {
      fprintf(stderr, "writing blm model: %s\n", 
              config["out-blm"].get_c_str()); 
      lm.write(io::Stream(config["out-blm"].get_str(), "w"));
    }
  }
  catch (std::string &str) {
    fprintf(stderr, "exception: %s\n", str.c_str());
    exit(1);
  }
  catch (std::exception &e) {
    fprintf(stderr, "exception: %s\n", e.what());
    exit(1);
  }
}

