#ifndef PROGRESS_HH
#define PROGRESS_HH

#include <math.h>
#include <string>
#include <stdio.h>
#include "Timer.hh"

/** Display progress string and compute. */
class Progress {
public:

  /** Create progress object.
   *
   * \param skip = number of counts to skip between each report
   * \param max_count = maximum count for estimating time left
   * (zero if not used at all)
   */
  Progress(long skip = 0, long max_count = 0)
  {
    m_file = stderr;
    m_count = 0;
    m_skip_count = 0;
    m_skip = skip;
    m_max_count = max_count;
    m_timer.start();
  }

  /** Set the FILE stream for printing the report. */
  void set_file(FILE *file)
  {
    m_file = file;
  }

  /** Set the report string shown before numerical report. */
  void set_report_string(std::string str)
  {
    m_report_str = str;
  }

  /** Reset timer for estimating time left. */
  void reset_timer()
  {
    m_timer.stop();
    m_timer.reset();
    m_timer.start();
  }

  /** Increase progress counter.
   * \param steps = steps to increment
   */
  void step(long steps = 1)
  {
    m_count += steps;
    m_skip_count -= steps;
    if (m_skip_count <= 0) {
      show_report();
      m_skip_count = m_skip + 1;
    }
  }

  /** Show the progress report. */
  void show_report()
  {
    if (m_max_count <= 0)
      fprintf(m_file, "\r%s %ld ", m_report_str.c_str(), m_count);
    else {
      m_timer.snapshot();
      float real_sec = m_timer.real_sec();
      float fraction = (float)m_count / m_max_count;
      
      if (real_sec < 1)
        real_sec = 1;

      int secs_left = lrintf(real_sec / fraction - real_sec);
      int mins_left = secs_left / 60;
      secs_left = secs_left % 60;
      int hours_left = mins_left / 60;
      mins_left = mins_left % 60;

      int secs_elapsed = (int)real_sec;
      int mins_elapsed = secs_elapsed / 60;
      secs_elapsed = secs_elapsed % 60;
      int hours_elapsed = mins_elapsed / 60;
      mins_elapsed = mins_elapsed % 60;

      fprintf(m_file, 
              "\r%s %ld / %ld (%.2f %%)  %02d:%02d:%02d  ETA %02d:%02d:%02d ", 
              m_report_str.c_str(), m_count, m_max_count,
              100.0 * fraction, 
              hours_elapsed, mins_elapsed, (int)secs_elapsed,
              hours_left, mins_left, secs_left);
    }
  }

  /** Finish report by printing newline and stopping timer. */
  void finish()
  {
    show_report();
    fputc('\n', m_file);
  }

private:
  
  /** FILE stream for printing the report. */
  FILE *m_file;

  /** Prefix string for the report (empty if not used) */
  std::string m_report_str;

  /** Current progress count. */
  long m_count;

  /** Decreasing counter for next report. */
  long m_skip_count;

  /** Number of counts to skip between showing the report string (0 =
   * report every time). */
  long m_skip;

  /** Counter value when the job should be finished.  Zero if not
   * used. */
  long m_max_count;

  /** Timer used for estimating time left (used only if \c m_max_count
   * is positive). */
  Timer m_timer;
};

#endif /* PROGRESS_HH */
